module FP.Prelude.Vector 
  ( module FP.Prelude.Vector
  , module Data.Vector.Unboxed
  ) where

import FP.Prelude.Core

import Data.Vector.Unboxed (Unbox)

import qualified Data.Vector         as Vector
import qualified Data.Vector.Unboxed as UVector

-- Boxed

type 𝕍 = Vector.Vector

instance Monoid (𝕍 a) where {null = Vector.empty;(⧺) = (Vector.++)}
instance ToStream a (𝕍 a) where stream = stream ∘ Vector.toList
instance ToFold a (𝕍 a) where fold = fold ∘ stream
instance Lookup ℕ a (𝕍 a) where lookup n v = v Vector.!? 𝕚 n

vec ∷ (ToFold a t) ⇒ t → 𝕍 a
vec = Vector.fromList ∘ list

-- Unboxed

type 𝕍ᵘ = UVector.Vector

instance (Unbox a) ⇒ Monoid (𝕍ᵘ a) where {null = UVector.empty;(⧺) = (UVector.++)}
instance (Unbox a) ⇒ ToStream a (𝕍ᵘ a) where stream = stream ∘ UVector.toList
instance (Unbox a) ⇒ ToFold a (𝕍ᵘ a) where fold = fold ∘ stream
instance (Unbox a) ⇒ Lookup ℕ a (𝕍ᵘ a) where lookup n v = v UVector.!? 𝕚 n

uvec ∷ (ToFold a t,Unbox a) ⇒ t → 𝕍ᵘ a
uvec = UVector.fromList ∘ list

-- Matrix

data 𝕄 a = Matrix 
  { matrixRowSize ∷ ℕ
  , matrixData ∷ 𝕍 a 
  }

instance Lookup (ℕ,ℕ) a (𝕄 a) where 
  lookup (i,j) (Matrix rowSize dat) = dat # (i × rowSize + j)

matrix ∷ (ToFold a t) ⇒ ℕ → t → 𝕄 a
matrix n = Matrix n ∘ vec

-- Unboxed Matrix

data 𝕄ᵘ a = UMatrix 
  { umatrixRowSize ∷ ℕ
  , umatrixData ∷ 𝕍ᵘ a
  }

instance (Unbox a) ⇒ Lookup (ℕ,ℕ) a (𝕄ᵘ a) where 
  lookup (i,j) (UMatrix rowSize dat) = dat # (i × rowSize + j)
  
umatrix ∷ (ToFold a t,Unbox a) ⇒ ℕ → t → 𝕄ᵘ a
umatrix n = UMatrix n ∘ uvec
